@@ -1,21 +1,22 @@
-/*
-    mafw-lastfm: a last.fm scrobbler for mafw
-    Copyright (C) 2009  Claudio Saavedra <csaavedra@igalia.com>
-
-    This program is free software; you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation; either version 2 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
-
-    You should have received a copy of the GNU General Public License
-    along with this program; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
-*/
+/**
+ * mafw-lastfm: a last.fm scrobbler for mafw
+ *
+ * Copyright (C) 2009  Claudio Saavedra <csaavedra@igalia.com>
+ *
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; either version 2 of the License, or
+ * (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software
+ * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ */
 
 #include <glib.h>
 #include <libsoup/soup.h>
@@ -26,7 +27,7 @@
 #define CLIENT_ID "maf"
 #define CLIENT_VERSION "0.0.1"
 
-G_DEFINE_TYPE (MafwLastfmScrobbler, mafw_lastfm_scrobbler, G_TYPE_OBJECT)
+G_DEFINE_TYPE (MafwLastfmScrobbler, mafw_lastfm_scrobbler, G_TYPE_OBJECT);
 
 #define GET_PRIVATE(o) \
   (G_TYPE_INSTANCE_GET_PRIVATE ((o), MAFW_LASTFM_TYPE_SCROBBLER, MafwLastfmScrobblerPrivate))
@@ -38,13 +39,16 @@ typedef enum {
   MAFW_LASTFM_SCROBBLER_SUBMITTING
 } MafwLastfmScrobblerStatus;
 
-struct _MafwLastfmScrobblerPrivate {
+struct MafwLastfmScrobblerPrivate {
   SoupSession *session;
   gchar *session_id;
   gchar *np_url;
   gchar *sub_url;
   GQueue *scrobbling_queue;
   guint handshake_id;
+  guint retry_id;
+  guint playing_now_id;
+  MafwLastfmTrack *playing_now_track;
 
   guint retry_interval;
   SoupMessage *retry_message;
@@ -62,76 +66,41 @@ static MafwLastfmTrack *
 mafw_lastfm_track_encode (MafwLastfmTrack *track);
 static gboolean mafw_lastfm_track_cmp (MafwLastfmTrack *a,
                                        MafwLastfmTrack *b);
+static  MafwLastfmTrack *
+mafw_lastfm_track_dup (MafwLastfmTrack *track);
 
 static void handshake_cb (SoupSession *session,
                           SoupMessage *message,
                           gpointer user_data);
 
 static void
-mafw_lastfm_scrobbler_get_property (GObject *object, guint property_id,
-                              GValue *value, GParamSpec *pspec)
-{
-  switch (property_id) {
-  default:
-    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
-  }
-}
-
-static void
-mafw_lastfm_scrobbler_set_property (GObject *object, guint property_id,
-                              const GValue *value, GParamSpec *pspec)
-{
-  switch (property_id) {
-  default:
-    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
-  }
-}
-
-static void
-mafw_lastfm_scrobbler_dispose (GObject *object)
+mafw_lastfm_scrobbler_finalize (GObject *object)
 {
   MafwLastfmScrobblerPrivate *priv = MAFW_LASTFM_SCROBBLER (object)->priv;
 
-  if (priv->session != NULL) {
+  if (priv->session) {
     soup_session_abort (priv->session);
     g_object_unref (priv->session);
-    priv->session = NULL;
   }
 
-  if (priv->session_id != NULL) {
-    g_free (priv->session_id);
-    priv->session_id = NULL;
-  }
-  if (priv->np_url == NULL) {
-    g_free (priv->np_url);
-    priv->np_url = NULL;
-  }
-
-  if (priv->sub_url == NULL) {
-    g_free (priv->sub_url);
-    priv->sub_url = NULL;
-  }
+  g_free (priv->session_id);
+  g_free (priv->np_url);
+  g_free (priv->sub_url);
 
   if (priv->scrobbling_queue) {
     g_queue_foreach (priv->scrobbling_queue, (GFunc) mafw_lastfm_track_free, NULL);
     g_queue_free (priv->scrobbling_queue);
-    priv->scrobbling_queue = NULL;
   }
 
-  if (priv->username) {
-    g_free (priv->username);
-  }
+  if (priv->playing_now_id)
+    g_source_remove (priv->playing_now_id);
 
-  if (priv->md5password) {
-    g_free (priv->md5password);
-  }
+  if (priv->playing_now_track)
+    mafw_lastfm_track_free (priv->playing_now_track);
 
-  G_OBJECT_CLASS (mafw_lastfm_scrobbler_parent_class)->dispose (object);
-}
+  g_free (priv->username);
+  g_free (priv->md5password);
 
-static void
-mafw_lastfm_scrobbler_finalize (GObject *object)
-{
   G_OBJECT_CLASS (mafw_lastfm_scrobbler_parent_class)->finalize (object);
 }
 
@@ -141,9 +110,6 @@ mafw_lastfm_scrobbler_class_init (MafwLastfmScrobblerClass *klass)
   GObjectClass *object_class = G_OBJECT_CLASS (klass);
   g_type_class_add_private (klass, sizeof (MafwLastfmScrobblerPrivate));
 
-  object_class->get_property = mafw_lastfm_scrobbler_get_property;
-  object_class->set_property = mafw_lastfm_scrobbler_set_property;
-  object_class->dispose = mafw_lastfm_scrobbler_dispose;
   object_class->finalize = mafw_lastfm_scrobbler_finalize;
 }
 
@@ -159,12 +125,16 @@ mafw_lastfm_scrobbler_init (MafwLastfmScrobbler *scrobbler)
   priv->sub_url = NULL;
   priv->scrobbling_queue = g_queue_new ();
   priv->handshake_id = 0;
+  priv->retry_id = 0;
 
   priv->retry_message = NULL;
   priv->retry_interval = 5;
   priv->scrobble_list = NULL;
   priv->suspended_track = NULL;
 
+  priv->playing_now_track = NULL;
+  priv->playing_now_id = 0;
+
   priv->username = NULL;
   priv->md5password = NULL;
 
@@ -182,15 +152,13 @@ mafw_lastfm_scrobbler_set_credentials (MafwLastfmScrobbler *scrobbler,
                                        const gchar *username,
                                        const gchar *md5password)
 {
-  if (scrobbler->priv->username) {
-    g_free (scrobbler->priv->username);
-  }
+  g_free (scrobbler->priv->username);
   scrobbler->priv->username = g_strdup (username);
 
-  if (scrobbler->priv->md5password) {
-    g_free (scrobbler->priv->md5password);
-  }
+  g_free (scrobbler->priv->md5password);
   scrobbler->priv->md5password = g_strdup (md5password);
+
+  scrobbler->priv->status = MAFW_LASTFM_SCROBBLER_NEED_HANDSHAKE;
 }
 
 static gboolean
@@ -222,7 +190,7 @@ mafw_lastfm_scrobbler_scrobbling_failed (MafwLastfmScrobbler *scrobbler)
   GList *iter;
 
   for (iter = g_list_last (scrobbler->priv->scrobble_list);
-       iter != NULL; iter = iter->prev) {
+       iter; iter = iter->prev) {
     g_queue_push_head (scrobbler->priv->scrobbling_queue,
                        iter->data);
   }
@@ -266,7 +234,7 @@ mafw_lastfm_scrobbler_scrobble_list (MafwLastfmScrobbler *scrobbler,
 
   post_data = g_strdup_printf ("s=%s", scrobbler->priv->session_id);
 
-  for (iter = list, i = 0; iter != NULL; iter = iter->next, i++) {
+  for (iter = list, i = 0; iter; iter = iter->next, i++) {
     track = (MafwLastfmTrack *) iter->data;
     track_data = g_strdup_printf ("&a[%i]=%s&t[%i]=%s&i[%i]=%li&o[%i]=%c&r[%i]=&l[%i]=%lld&b[%i]=%s&n[%i]=%i&m[%i]=",
                                   i, track->artist,
@@ -304,33 +272,29 @@ set_playing_now_cb (SoupSession *session,
 
   if (SOUP_STATUS_IS_SUCCESSFUL (message->status_code)) {
     g_print ("Playing-now: %s", message->response_body->data);
-    if (strcmp (message->response_body->data, "BADSESSION\n") == 0) {
+    if (strcmp (message->response_body->data, "BADSESSION\n") == 0)
       mafw_lastfm_scrobbler_defer_handshake (scrobbler);
-    }
   }
 }
 
 void
 mafw_lastfm_scrobbler_set_playing_now (MafwLastfmScrobbler *scrobbler,
-                                       MafwLastfmTrack     *track)
+                                       MafwLastfmTrack *encoded)
 {
   gchar *post_data;
   SoupMessage *message;
-  MafwLastfmTrack *encoded;
 
   g_return_if_fail (MAFW_LASTFM_IS_SCROBBLER (scrobbler));
-  g_return_if_fail (track != NULL);
+  g_return_if_fail (encoded);
   g_return_if_fail (scrobbler->priv->status == MAFW_LASTFM_SCROBBLER_READY);
 
-  encoded = mafw_lastfm_track_encode (track);
-  post_data = g_strdup_printf ("s=%s&a=%s&t=%s&b=%s&l=%lli&n=%u&m=",
+  post_data = g_strdup_printf ("s=%s&a=%s&t=%s&b=%s&l=%lli&n=%i&m=",
                                scrobbler->priv->session_id,
                                encoded->artist,
                                encoded->title,
-			       encoded->album ? encoded->album : "",
+                               encoded->album ? encoded->album : "",
                                encoded->length,
                                encoded->number);
-  g_free (encoded);
 
   message = soup_message_new ("POST",
                               scrobbler->priv->np_url);
@@ -352,14 +316,13 @@ scrobble_real (MafwLastfmScrobbler *scrobbler)
   GList *list = NULL;
   gint tracks = 0;
 
-  if (scrobbler->priv->status != MAFW_LASTFM_SCROBBLER_READY) {
+  if (scrobbler->priv->status != MAFW_LASTFM_SCROBBLER_READY)
     return;
-  }
 
   while (!g_queue_is_empty (scrobbler->priv->scrobbling_queue) && tracks < 50) {
     track = g_queue_pop_head (scrobbler->priv->scrobbling_queue);
     list = g_list_append (list, track);
-    tracks ++;
+    tracks++;
   }
 
   if (list) {
@@ -403,13 +366,25 @@ void
 mafw_lastfm_scrobbler_suspend (MafwLastfmScrobbler *scrobbler)
 {
   /* Nothing to suspend. */
-  if (scrobbler->priv->suspended_track != NULL)
+  if (scrobbler->priv->suspended_track)
     return;
 
   /* Remove the last track from the queue, since it is suspended. */
   scrobbler->priv->suspended_track = g_queue_pop_tail (scrobbler->priv->scrobbling_queue);
 }
 
+static gboolean
+defer_set_playing_now_cb (MafwLastfmScrobbler *scrobbler)
+{
+  mafw_lastfm_scrobbler_set_playing_now (scrobbler,
+                                         scrobbler->priv->playing_now_track);
+  mafw_lastfm_track_free (scrobbler->priv->playing_now_track);
+  scrobbler->priv->playing_now_track = NULL;
+  scrobbler->priv->playing_now_id = 0;
+
+  return FALSE;
+}
+
 void
 mafw_lastfm_scrobbler_enqueue_scrobble (MafwLastfmScrobbler *scrobbler,
                                         MafwLastfmTrack *track)
@@ -418,12 +393,20 @@ mafw_lastfm_scrobbler_enqueue_scrobble (MafwLastfmScrobbler *scrobbler,
 
   mafw_lastfm_scrobbler_flush_queue (scrobbler);
 
-  if (scrobbler->priv->status == MAFW_LASTFM_SCROBBLER_READY) {
-    mafw_lastfm_scrobbler_set_playing_now (scrobbler, track);
-  }
-
   encoded = mafw_lastfm_track_encode (track);
 
+  if (scrobbler->priv->status == MAFW_LASTFM_SCROBBLER_READY) {
+    if (scrobbler->priv->playing_now_id)
+      g_source_remove (scrobbler->priv->playing_now_id);
+    if (scrobbler->priv->playing_now_track)
+      mafw_lastfm_track_free (scrobbler->priv->playing_now_track);
+
+    scrobbler->priv->playing_now_track = mafw_lastfm_track_dup (encoded);
+    scrobbler->priv->playing_now_id = g_timeout_add_seconds (3,
+                                                             (GSourceFunc) defer_set_playing_now_cb,
+                                                             scrobbler);
+  }
+
   if (scrobbler->priv->suspended_track) {
     if (mafw_lastfm_track_cmp (scrobbler->priv->suspended_track,
                                encoded)) {
@@ -457,7 +440,7 @@ get_auth_string (const gchar *md5passwd,
   gchar *auth;
   gchar *md5;
 
-  g_return_val_if_fail (timestamp != NULL, NULL);
+  g_return_val_if_fail (timestamp, NULL);
 
   g_get_current_time (&time_val);
 
@@ -471,37 +454,51 @@ get_auth_string (const gchar *md5passwd,
   return md5;
 }
 
-static gboolean
+typedef enum {
+  AS_RESPONSE_OK,
+  /* AS_RESPONSE_BANNED, */
+  /* AS_RESPONSE_BADAUTH, */
+  AS_RESPONSE_BADTIME,
+  /* AS_RESPONSE_FAILED, */
+  AS_RESPONSE_OTHER
+} AsHandshakeResponse;
+
+static AsHandshakeResponse
 parse_handshake_response (MafwLastfmScrobbler *scrobbler,
                           const gchar *response_data)
 {
   gchar **response;
-
+  AsHandshakeResponse retval;
   response = g_strsplit (response_data, "\n", 5);
 
   if (g_str_has_prefix (response [0], "OK")) {
-
     g_free (scrobbler->priv->session_id);
     g_free (scrobbler->priv->np_url);
     g_free (scrobbler->priv->sub_url);
 
-    scrobbler->priv->session_id = response [1];
+    scrobbler->priv->session_id = response[1];
     scrobbler->priv->np_url = response[2];
     scrobbler->priv->sub_url = response[3];
 
     /* We take ownership on the relevant parsed data, free the
        array and response code. */
-    g_free (response [0]);
-    g_free (response [4]);
+    g_free (response[0]);
+    g_free (response[4]);
     g_free (response);
 
-    return TRUE;
+    retval = AS_RESPONSE_OK;
+  } else if (g_str_has_prefix (response [0], "BADTIME")) {
+    retval = AS_RESPONSE_BADTIME;
   } else {
-    g_warning ("Couldn't handshake: %s", response [0]);
-    g_strfreev (response);
-
-    return FALSE;
+    retval = AS_RESPONSE_OTHER;
   }
+
+  if (retval != AS_RESPONSE_OK) {
+    g_warning ("Couldn't handshake: %s", response[0]);
+    g_strfreev (response);
+  }
+
+  return retval;
 }
 
 static gboolean
@@ -513,22 +510,32 @@ retry_queue_message (gpointer userdata)
                               priv->retry_message,
                               handshake_cb,
                               MAFW_LASTFM_SCROBBLER (userdata));
+  priv->retry_message = NULL;
+
   return FALSE;
 }
 
 static void
 handshake_cb (SoupSession *session,
-               SoupMessage *message,
-               gpointer user_data)
+              SoupMessage *message,
+              gpointer user_data)
 {
   MafwLastfmScrobbler *scrobbler = MAFW_LASTFM_SCROBBLER (user_data);
 
   if (SOUP_STATUS_IS_SUCCESSFUL (message->status_code)) {
     g_print ("%s", message->response_body->data);
-    if (parse_handshake_response (scrobbler, message->response_body->data)) {
+    switch (parse_handshake_response (scrobbler, message->response_body->data)) {
+    case AS_RESPONSE_OK:
       scrobbler->priv->status = MAFW_LASTFM_SCROBBLER_READY;
       scrobbler->priv->retry_interval = 5;
       return;
+    case AS_RESPONSE_BADTIME:
+      scrobbler->priv->status = MAFW_LASTFM_SCROBBLER_NEED_HANDSHAKE;
+      scrobbler->priv->retry_interval = 5;
+      mafw_lastfm_scrobbler_handshake (scrobbler);
+      return;
+    case AS_RESPONSE_OTHER:
+      break;
     }
   }
 
@@ -536,9 +543,9 @@ handshake_cb (SoupSession *session,
   g_print ("message failed, trying to send in %d seconds.\n", scrobbler->priv->retry_interval);
   scrobbler->priv->status = MAFW_LASTFM_SCROBBLER_NEED_HANDSHAKE;
   scrobbler->priv->retry_message = g_object_ref (message);
-  g_timeout_add_seconds (scrobbler->priv->retry_interval,
-                         retry_queue_message,
-                         scrobbler);
+  scrobbler->priv->retry_id = g_timeout_add_seconds (scrobbler->priv->retry_interval,
+                                                     retry_queue_message,
+                                                     scrobbler);
   if (scrobbler->priv->retry_interval < 320)
     scrobbler->priv->retry_interval *= 2;
 }
@@ -546,31 +553,42 @@ handshake_cb (SoupSession *session,
 void
 mafw_lastfm_scrobbler_handshake (MafwLastfmScrobbler *scrobbler)
 {
-        gchar *auth;
-        glong timestamp;
-        gchar *handshake_url;
-        SoupMessage *message;
+  gchar *auth;
+  glong timestamp;
+  gchar *handshake_url;
+  SoupMessage *message;
 
-        g_return_if_fail (scrobbler->priv->status != MAFW_LASTFM_SCROBBLER_HANDSHAKING);
-        g_return_if_fail (scrobbler->priv->username != NULL || scrobbler->priv->md5password != NULL);
+  g_return_if_fail (scrobbler->priv->status != MAFW_LASTFM_SCROBBLER_HANDSHAKING);
+  g_return_if_fail (scrobbler->priv->username || scrobbler->priv->md5password);
+  if (scrobbler->priv->retry_id) {
+    g_source_remove (scrobbler->priv->retry_id);
+    scrobbler->priv->retry_id = 0;
+    if (scrobbler->priv->retry_message) {
+      g_object_unref (scrobbler->priv->retry_message);
+      scrobbler->priv->retry_message = NULL;
+    }
+  }
+  if (scrobbler->priv->handshake_id) {
+    g_source_remove (scrobbler->priv->handshake_id);
+    scrobbler->priv->handshake_id = 0;
+  }
+  scrobbler->priv->status = MAFW_LASTFM_SCROBBLER_HANDSHAKING;
 
-        scrobbler->priv->status = MAFW_LASTFM_SCROBBLER_HANDSHAKING;
+  auth = get_auth_string (scrobbler->priv->md5password, &timestamp);
 
-        auth = get_auth_string (scrobbler->priv->md5password, &timestamp);
+  handshake_url = g_strdup_printf ("http://post.audioscrobbler.com/?hs=true&p=1.2.1&c=%s&v=%s&u=%s&t=%li&a=%s",
+                                   CLIENT_ID, CLIENT_VERSION,
+                                   scrobbler->priv->username,
+                                   timestamp,
+                                   auth);
 
-        handshake_url = g_strdup_printf ("http://post.audioscrobbler.com/?hs=true&p=1.2.1&c=%s&v=%s&u=%s&t=%li&a=%s",
-                                         CLIENT_ID, CLIENT_VERSION,
-                                         scrobbler->priv->username,
-                                         timestamp,
-                                         auth);
-
-        message = soup_message_new ("GET", handshake_url);
-        soup_session_queue_message (scrobbler->priv->session,
-                                    message,
-                                    handshake_cb,
-                                    scrobbler);
-        g_free (handshake_url);
-        g_free (auth);
+  message = soup_message_new ("GET", handshake_url);
+  soup_session_queue_message (scrobbler->priv->session,
+                              message,
+                              handshake_cb,
+                              scrobbler);
+  g_free (handshake_url);
+  g_free (auth);
 }
 
 MafwLastfmTrack *
@@ -582,7 +600,7 @@ mafw_lastfm_track_new (void)
 void
 mafw_lastfm_track_free (MafwLastfmTrack *track)
 {
-  if (track == NULL)
+  if (!track)
     return;
 
   g_free (track->artist);
@@ -601,13 +619,13 @@ mafw_lastfm_track_encode (MafwLastfmTrack *track)
 
   encoded = mafw_lastfm_track_new ();
 
-  if (track->artist != NULL)
+  if (track->artist)
     encoded->artist = soup_uri_encode (track->artist, EXTRA_URI_ENCODE_CHARS);
 
-  if (track->title != NULL)
+  if (track->title)
     encoded->title = soup_uri_encode (track->title, EXTRA_URI_ENCODE_CHARS);
 
-  if (track->album != NULL)
+  if (track->album)
     encoded->album = soup_uri_encode (track->album, EXTRA_URI_ENCODE_CHARS);
 
   encoded->length = track->length;
@@ -628,3 +646,19 @@ mafw_lastfm_track_cmp (MafwLastfmTrack *a,
           (!(a->album || b->album) ||
            ((a->album && b->album) && strcmp (a->album, b->album) == 0)));
 }
+
+static MafwLastfmTrack *
+mafw_lastfm_track_dup (MafwLastfmTrack *track)
+{
+  MafwLastfmTrack *track2 = mafw_lastfm_track_new ();
+
+  track2->artist = g_strdup (track->artist);
+  track2->title = g_strdup (track->title);
+  track2->album = g_strdup (track->album);
+  track2->timestamp = track->timestamp;
+  track2->source = track->source;
+  track2->length = track->length;
+  track2->number = track->number;
+
+  return track2;
+}
